(*****************************************************************************)
(* "Experiments 1" contains implementation of functions for

   C. Bodei, M. Buchholtz, P. Degano, H. Riis Nielson, F. Nielson: 
   "Automatic Validation of Protocol Narration", Proc. of CSFW'03,
   IEEE 2003.							

   This file contains main main functions to run the experiments.

                                                22.12.2003. Mikael Buchholtz *)
(*****************************************************************************)


Compiler.Control.Print.printDepth  := 4000;
Compiler.Control.Print.stringDepth := 1000;
Compiler.Control.Print.printLength := 40;

CM.make();


local
    open AuxEx1
 
    structure SS = FormulaAnalyzer
			    
    fun between []      e = []
      | between [t]     e = [t]
      | between (l::ls) e = l::e::between ls e

    val dl = [((fn x=>x),"$A \\neq B$ and $K_A \\neq K_B$"),
	      (chooseRoleABS,"$A = B$ and $K_A \\neq K_B$"),
	      (chooseRoleKey,"$A \\neq B$ and $K_A = K_B$"),

	      (chooseRoleBoth,"$A = B$ and $K_A = K_B$")]	


    (* s2tex: string to LaTex -- a very messy function *)
    (* A better parser to the Succinct Solver would eradicate the need for it *)
    fun s2tex s =

	(* Some functions on strings *)

	let fun splitl(s, c) = 
		let val sl = String.explode s 				 
			     
		    fun findSplit(f,[],c) = (f,[])
		      | findSplit(f,l,c) = if c = List.hd l 
					   then (f,l) 
					   else findSplit(f@[List.hd l], List.tl l,c)
						
		    val (fl,ll) = findSplit([],sl,c)
		in 
		    (String.implode(fl),String.implode(ll))
		end
			
	    fun String_tl(s) = if s = "" 
			       then s 
			       else String.implode(List.tl(String.explode(s)))
					 
					 
	    (* String.substring acts strange -- at least I don't understand it.
				Here is my own implementation instead *)
	    fun substring(s, i, j) = String.implode			  
					 (List.drop(List.take(String.explode(s),j), i))
					 
	    (* isPrefix is strangely enough NOT implemented in the String libary of SML/NJ *)
					 
	    fun isPrefix ("") (_)  = true
	      | isPrefix (ps) ("") = false
	      | isPrefix (ps) (s)  = if (String.size s >= String.size ps) 
				     then (ps = substring(s, 0, String.size ps))
				     else false
					
  						  
	    fun toStringName(s) = let val (s1,s2) = splitl(s,#"_")
				  in
				      "{"^s1^"_{"^
				      (String.implode
					   (List.map 
						(fn c => if c = #"_" then #"-" else c) 
						(String.explode(String_tl(s2)))))
				      ^"}}"
				  end
				      
	in		
	    if      s = LLysa.CPDY
            then "\\ell_\\bullet"
						     
	    else if s = Analysis1.canon(Analysis1.NDY)
	    then "n_\\bullet"

	    else if s = Analysis1.canon(Analysis1.NPDY)  
            then "n_\\bullet^+"

	    else if s = Analysis1.canon(Analysis1.NMDY) 
            then "n_\\bullet^-"

	    else if s = "ALL_FROM_DY" 
            then "\\rho(z_\\bullet)"
			  
	    else if (isPrefix "_PLUS_" s) 
            then toStringName(substring(s,6,String.size s))^"^+"

	    else if (isPrefix "_MINUS_" s) 
            then toStringName(substring(s,7,String.size s))^"^-" 

	    else toStringName(s)
	end					
    

    (* rel2tex : maps s2tex onto elements from a relation and rows generates
    in a LaTeX table                                                         *)

    fun  rel2tex(rl) =
	 String.concat(between
	      (List.map 
		   (fn el => "("^String.concat(between
				      (List.map 
					   (fn e => "$"^s2tex(e)^"$") el)
				      ", ")^")") rl)
	      ", ")
	
    fun res2tex(uni, []) = ""
      | res2tex(uni, (name,rl)::ls) = rel2tex(rl)^
				      "\n\n"^res2tex(uni,ls)			      
    

in

(*****************************************************************************)
(*                                                                           *)
(* solve(prot, n, l, d)  main function for running the analysis.             *)
(*                                                                           *)
(* prot    Protocol to be analysed                                           *)
(* n       Number of unfoldings                                              *)
(* l       Leak number.  0 indicates no leak.                                *)
(* d       Distinction - 0 indicates roles and keys distinct                 *)
(*                       1 indicates roles identical but keys distinct       *)
(*                       2 indicates roles distinct but keys identical       *)
(*                       3 indicates roles and keys identical                *)
(*                                                                           *)
(*****************************************************************************)
(* Protocols, prot,  are on the form (lab, name, p, leak, obs) where         *)
(*                                                                           *)
(* lab  : String    is a label (used e.g. in file names)                     *)
(* name : String    is the name of the protocol                              *)
(* p    : Lysa.Proc * Lysa.Proc * Lysa.Proc -> Lysa.Proc                     *)
(*                  is a mapping from three "leaks" to a Meta LySa process   *)
 (* leak : ...       is a list of leaks                                       *)
(* obs  : ...       is a list of observation variables at Meta LySa level    *)
(*****************************************************************************)


fun solve (prot, n, l, d) = 
    let

    (********************************************************************************)
    
    val (lab,text,p,leak,obs) = prot


    (* Choose leak, distinct  *)

    val prot1  = chooseLeak l prot
    val (chooseRole,gs) = List.nth(dl,d)  
    val prot2  = chooseRole(prot1)


    (* Generate Lysa process with n instances *)
    
    val prot3  = ml2lAll n prot2    
    val (_,_,lp,lobs) = prot3


    (* Generate Labelled Lysa process *)
    
    val _      = Lysa2LLysa.reset()
    val llp    = Lysa2LLysa.l2llProc(lp)		  
    val uol    = List.map Lysa2LLysa.l2llTerm lobs
 

    (* Run analysis *)

    val _      = TextIO.print(text^"\n\n")
    val clauseStr = Analysis1.analyse(llp)^"&\n\n"^
		 "/****************************************************************/\n"^
		 "/* Observation predicates                                       */\n"^
		 "/****************************************************************/\n\n"^
		 "("^Analysis1.genObs(llp,uol)^") \n"

    val fi     = "results/last_analysis.cl"
    val fh     = TextIO.openOut(fi)	
    val _      = TextIO.output(fh, clauseStr)
    val _      = TextIO.closeOut(fh)	
	
    val name   = lab^"_"^Int.toString(n)^"_"^Int.toString(l)^"_"^Int.toString(d)
    val ofile  = "results/"^name
		
    val solver = SS.init()
    val clause = SS.inputStr(clauseStr)	
    val _      = SS.solve(solver, clause)

    val str    = SS.outputFile(solver, ofile^".rel") ([], true)


    (* Create lots of files *)

    val s      = "\\begin{tabular}{|p{4cm}|p{4cm}|p{5cm}|}\n"^
		 "\\hline\n"^
		 res2tex(SS.outputData solver ([("PSI",2)], false))^" &\n"^
		 res2tex(SS.outputData solver ([("DYN",1)], false))^" &\n"^
		 res2tex(SS.outputData solver ([("OBS_VAR",2)], false))^" \\\\ \n"^
		 "\\hline\n"^
		 "\\end{tabular}\n"
	    

    val (ls,_) = if l = 0 
		 then ("No leaks",(Lysa.NIL,Lysa.NIL,Lysa.NIL)) 
		 else List.nth(leak,(l-1))

	    
    val st     = "\\begin{table}\n"^
		 "\\begin{tabular}{l}\n"^
		 "\\begin{tabular}{|p{4cm}|p{4cm}|p{5cm}|}\n"^		 
		 "\\hline\n"^
		 "$\\psi$ & $\\rho(z_\\bullet)\\cap{\\cal N}$ & Observation variables\\\\\n "^
		 "\\end{tabular}\n"^
		 "\\\\\n"^
		 "\\input{"^name^".tex}\n"^		 
		 "\\end{tabular}\n"^
		 "\\caption{\\label{tab:"^name^"}"^
		 "\\textbf{"^text^"},\n"^
		 ls^",\n"^
		 gs^",\n"^
		 "$n="^Int.toString(n)^"$\n"^
		 "}\n"^
		 "\\end{table}"
	     
    val fh     = TextIO.openOut(ofile^".tex")	
    val _      = TextIO.output(fh, s)
    val _      = TextIO.closeOut(fh)
	     
    val fh     = TextIO.openOut(ofile^"_table.tex")	
    val _      = TextIO.output(fh, st)
    val _      = TextIO.closeOut(fh)
		 
    val fh     = TextIO.openOut("results/last_analysis.tex")	
    val _      = TextIO.output(fh, st)
    val _      = TextIO.closeOut(fh)
    
    in
	()
    end



(************************************************************************************)
(* rs: re-solves the clauses in last_analysis.cl and latex result.tex with the      *)
(* result                                                                           *)
(************************************************************************************)

fun rs () =
    let
		
    val solver = SS.init()
    val clause = SS.inputFile("results/last_analysis.cl")	
    val _      = SS.solveCount(solver, clause)

    val str    = SS.outputFile(solver, "results/last_analysis.rel") ([], true)


    (* Create lots of files *)

    val s      = "\\begin{tabular}{|p{4cm}|p{4cm}|p{5cm}|}\n"^
		 "\\hline\n"^
		 res2tex(SS.outputData solver ([("PSI",2)], false))^" &\n"^
		 res2tex(SS.outputData solver ([("DYN",1)], false))^" &\n"^
		 res2tex(SS.outputData solver ([("OBS_VAR",2)], false))^" \\\\ \n"^
		 "\\hline\n"^
		 "\\end{tabular}\n"
		 

	    
    val st     = "\\begin{table}\n"^
		 "\\begin{tabular}{l}\n"^
		 "\\begin{tabular}{|p{4cm}|p{4cm}|p{5cm}|}\n"^		 
		 "\\hline\n"^
		 "$\\psi$ & $\\rho(z_\\bullet)\\cap{\\cal N}$ & Observation variables\\\\\n "^
		 "\\end{tabular}\n"^
		 "\\\\\n"^
		 s^
		 "\\end{tabular}\n"^
		 "\\caption{Solved 'results/last\\_analysis.sml'"^"}\n"^
		 "\\end{table}"
	     
	     
		 
    val fh     = TextIO.openOut("results/last_analysis.tex")	
    val _      = TextIO.output(fh, st)
    val _      = TextIO.closeOut(fh)
		 
    val _      = OS.Process.system("cd results; latex result.tex; "^
				   "dvips -o result.ps result.dvi")

    in
	()
    end



(*****************************************************************************)
(* so : Prot -> ()                                                           *)
(* Runs solve(prot,3,0,0)                                                    *)
(*****************************************************************************)

fun so(prot)  = solve(prot,2,0,0);
       

(*****************************************************************************)
(* sol : Prot -> ()                                                          *)
(* Runs solve(prot,3,0,0) and LaTeXes the file "results/result.tex" which    *)
(* contains an overview of the analysis result                               *)
(*****************************************************************************)

fun sol(prot) = (solve(prot,2,0,0);
              OS.Process.system("cd results; latex result.tex; "^
				"dvips -o result.ps result.dvi"))


(*****************************************************************************)
(* distinct: Prot -> ()                                                      *)
(* runs analyse of the protocol with various distinctions of roles and       *)
(* generates a LaTeX overview in the file "results/lab_distinct.tex"         *)
(*****************************************************************************)

fun distinct(p) = 
  
    let 
	val (lab,text,_,al,_) = p
	val n = 2
	val dcl = [0,1,2,3]		  
		  
	fun doSolve l d  = 
	    let val _ = solve(p,n,l,d)
	    in "\\input{results/"^lab^"_"^
	       Int.toString(n)^"_"^
	       Int.toString(l)^"_"^
	       Int.toString(d)^".tex}\\\\\n"
	    end

	val s  = "\\begin{table}\n"^
		 "\\begin{tabular}{l}\n"^
		 "\\begin{tabular}{|p{4cm}|p{4cm}|p{5cm}|}\n"^		 
		 "\\hline\n"^
		 "$\\psi$ & $\\rho(z_\\bullet)\\cap{\\cal N}$ & Observation variables\\\\\n "^
		 "\\hline\n"^
		 "\\end{tabular}\n"^
		 "\\\\\n"^		 
		 String.concat(List.map
			       (fn d => let val (_,ds) = List.nth(dl,d)
					in ds^"\\\\\n"^
					   (doSolve 0 d)
					end)
			       dcl)^		 
		 "\\end{tabular}\n"^
		 "\\caption{\\label{tab:"^lab^"_distinct}"^
		 "Distinct roles and keys for \\textbf{"^text^"} ("^lab^")}"^
		 "\\end{table}\n"
		
	val fh = TextIO.openOut("results/"^lab^"_distinct.tex")
	val _  = TextIO.output(fh, s)
	val _  = TextIO.closeOut(fh)	
		     
    in 
      ()
    end


(*****************************************************************************)
(* leak: Prot -> ()                                                          *)
(* runs analyse on all leaks in the protocol and generates a LaTeX overview  *)
(* in the file "results/lab_leak.tex"                                        *)
(*****************************************************************************)

fun leak(p) = 
    
    let 
	val (lab,text,_,al,_) = p
	val n = 2	  
	
	(* Generates list [0,1,...,length al] *)	
	val lcl = List.tabulate(((List.length al)+1),(fn x => x)) 
		  
	fun doSolve l d  = 
	    let val _ = solve(p,n,l,d)		
		val (ls,_) = if l = 0 
			     then ("No leaks",(Lysa.NIL,Lysa.NIL,Lysa.NIL)) 
			     else List.nth(al,(l-1))

	    in (ls, 
		"\\input{results/"^lab^"_"^
		Int.toString(n)^"_"^
		Int.toString(l)^"_"^
		Int.toString(d)^".tex}\\\\\n"
		)
	    end

	val s  = "\\begin{table}\n"^
		 "\\begin{tabular}{l}\n"^
		 "\\begin{tabular}{|p{4cm}|p{4cm}|p{5cm}|}\n"^		 
		 "\\hline\n"^
		 "$\\psi$ & $\\rho(z_\\bullet)\\cap{\\cal N}$ & Observation variables\\\\\n "^
		 "\\hline\n"^
		 "\\end{tabular}\n"^
		 "\\\\\n"^
		 String.concat(List.map (fn l => let val (ls, is) =  (doSolve l 0)
						 in
						     ls^"\\\\\n"^is
						 end)
					lcl)^
		 "\\end{tabular}\n"^
		 "\\caption{\\label{tab:"^lab^"_leak}"^
		 " Leaks for \\textbf{"^text^"} ("^lab^")}"^
		 "\\end{table}\n"
		
	val fh = TextIO.openOut("results/"^lab^"_leak.tex")
	val _  = TextIO.output(fh, s)
	val _  = TextIO.closeOut(fh)		     
    in 
      ()
    end


(*****************************************************************************)
(* analyseProtocol : Prot -> ()                                              *)
(* Runs prettyPrint, distinct, and leak on the protocol                      *)
(*****************************************************************************)

fun analyseProtocol p = (prettyPrint(p); distinct(p); leak(p))


(*****************************************************************************)
(* genDocument: FileName * Prot list                                         *)
(* Generates a LaTex document "tex/Filename.tex" that contains the result of *)
(* running the analysis of the protocols in Prot list                        *)
(*****************************************************************************)

fun genDocument dn pl =
    let 
	val s =  "\\documentclass{scrartcl}\n"^
		 "\\parindent 0ex\n"^
		 "\\parskip 2ex\n"^
		 "\\usepackage{color,float}\n"^
		 "\\input{macro.tex}\n"^		 
		 "\\begin{document}\n"^
		 "\\floatplacement{table}{!htbp}\n"^
		 String.concat(List.map 
			       (fn p => let val (lab,text,_,_,_) = p
					in
					    "\\section*{"^text^"}\n\n"^
					    "\\input{results/"^lab^".tex}\n\n"^
					    "\\input{results/"^lab^"_distinct.tex}\n\n"^
					    "\\input{results/"^lab^"_leak.tex}\n\n"^
					    "\\clearpage\n"
					end)
			       pl)^		 
		 "\\end{document}"		
	
	val fh = TextIO.openOut("tex/"^dn^".tex")
	val _  = TextIO.output(fh, s)
	val _  = TextIO.closeOut(fh)		     
    in
	()
    end    
	

(*****************************************************************************)
(* analyseList: FileName * Prot list                                         *)
(* As 'genDocument' but runs the analysis on the protocols as well.          *)
(* running the analysis of the protocols in Prot list                        *)
(*****************************************************************************)

fun analyseList dn pl = 
    ((genDocument dn pl); (List.map (analyseProtocol) pl))
end


(*****************************************************************************)
(* Below is some code for testing purposes that may be removed in a later    *)
(* version                                                                   *)
(*****************************************************************************)

fun pl (prot as (lab,text,_,_,_)) =    
    let 
	val s  = "\\documentclass{scrartcl}\n"^
		 "\\usepackage{color,float}\n"^
		 "\\input{../tex/macro.tex}\n"^
		 "\\begin{document}\n"^
		 "\\floatplacement{table}{!htbp}\n"^
		 "\\input{"^lab^".tex}\n"^
		 "\\end{document}"

	val fh = TextIO.openOut("results/pretty.tex")
	val _  = TextIO.output(fh, s)
	val _  = TextIO.closeOut(fh)
	val _  = AuxEx1.prettyPrint(prot)
    in
	(OS.Process.system("cd results; latex pretty.tex; "^
		       "dvips -o pretty.ps pretty.dvi"))
    end

(*
val catalysa = [WMF.wmf1,WMF.wmf2,WMF.wmf3,WMF.wmf6,
		Otway.or1,
		Yahalom.y1,Yahalom.y2,Yahalom.y3,Yahalom.y4,
		Andrew.a1,Andrew.a2,
		NS.ns1, NS.ns2, NS.ns3, NS.ns4]
*)
