(*****************************************************************************)
(* MSR protocol (Modular Square Root)                                        *)
(*                                                                           *)
(* M. J. Beller, Li-Fung Chang, and  Y. Yacobi: Privacy and Authentication   *)
(* on a Protable Communications System. IEEE Journal of Selected Areas in    *)
(* Communications. 11(6). p. 821-829. 1993.                                  *)
(*                                                                           *)
(* MSR, and IMSR protocols                                                   *)
(*                                               29.09.2003 Mikael Buchholtz *)
(*****************************************************************************)

structure MSR =
struct
local
    open Lysa
    open Protocols
in

val CA    = NAME("CA",["i"])
val Ks    = NAME("K",["i"])
val KUP   = NAMEP("K^U",[])
val KUM   = NAMEM("K^U",[])


val xKBs  = VAR("xKB",["i"])
val xCSs  = VAR("xCS",["i"])

val LAs1  = ("A1",["i"])
val LAs2  = ("A2",["i"])
val LAs3  = ("A3",["i"])
val LAs4  = ("A4",["i"])

val KSP   = NAMEP("K^S",[])
val KSM   = NAMEM("K^S",[])
val CS    = NAME("CS",[])


val zEK   = VAR("z1",["i"])
val zK    = VAR("z2",["i"])
val zEC   = VAR("z3",["i"])
val zC    = VAR("z4",["i"])

val LS1   = ("S1",[])
val LS2   = ("S2",[])
val LS3   = ("S3",[])
val LS4   = ("S4",[])



fun porig l = SOME (ORIG [l,CPDY]);
fun pdest l = SOME (DEST [l,CPDY]);

(*****************************************************************************)
val p1 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,			      
			    INP([PS,S,PA,A,PS,S],[xKBs],
			    NEW(Ks,
                            OUT([PA,A,PS,S,AENC([Ks],xKBs,LAs1,dest LS1)],
			    OUT([PA,A,PS,S,ENC([PA,A,CA],Ks,LAs2,dest LS2)],
				NIL)))))

                val BProc = NIL

                val SProc = PAR_I("i",1,
                            OUT([PS,S,PA,A,PS,S,KSP],
			    INP([PA,A,PS,S],[zEK],
                            ADEC(zEK,[],[zK],KSM,LS1,porig(LAs1),
                            INP([PA,A,PS,S],[zEC],
			    DEC(zEC,[PA,A,CA],[],zK,LS2,orig LAs2,
				 NIL))))))

             in	NEW_I([([("i",1)],CA)],
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)]))
             end
    in ("msr1",
	"MSR (n portables $I_i$, one Radio Control Equipment (RCE) $I_{-1}$)",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1)],xKBs),
	 ([("i",1)],zK)	 
	 ])
	   
    end 
(*****************************************************************************)



(*****************************************************************************)
val p2 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,			      
			    INP([PS,S,PA,A,PS,S],[xKBs],
			    NEW(Ks,
                            OUT([PA,A,PS,S,AENC([Ks],xKBs,LAs1,dest LS1)],
			    OUT([PA,A,PS,S,ENC([PA,A,
					        AENC([PA,A],KUM,LAs3, pdest LS3)],
					       Ks,LAs2,dest LS2)],
				NIL)))))

                val BProc = OUT([KUP,KSP,A,A,A],NIL)

                val SProc = PAR_I("i",1,
                            OUT([PS,S,PA,A,PS,S,KSP],
			    INP([PA,A,PS,S],[zEK],
                            ADEC(zEK,[],[zK],KSM,LS1,porig(LAs1),
                            INP([PA,A,PS,S],[zEC],
			    DEC(zEC,[PA,A],[zC],zK,LS2,orig LAs2,
		            ADEC(zC,[PA,A],[],KUP,LS3,dest LAs3,
				 NIL)))))))

             in	ANEW(NAME("K^U",[]),
                ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)]))
             end
    in ("msr2",
	"MSR (n portables $I_i$, one Radio Control Equipment (RCE) $I_{-1}$). "^
	"Certificate as signature.",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1)],xKBs),
	 ([("i",1)],zK)	 
	 ])
	   
    end 
(*****************************************************************************)



(*****************************************************************************)
val p3 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,			      
			    INP([PS,S,PA,A,PS,S,CS],[xKBs],
			    NEW(Ks,
                            OUT([PA,A,PS,S,AENC([Ks],xKBs,LAs1,dest LS1)],
			    OUT([PA,A,PS,S,ENC([PA,A,CA],Ks,LAs2,dest LS2)],
				NIL)))))

                val BProc = NIL

                val SProc = PAR_I("i",1,
                            OUT([PS,S,PA,A,PS,S,CS,KSP],
			    INP([PA,A,PS,S],[zEK],
                            ADEC(zEK,[],[zK],KSM,LS1,porig(LAs1),
                            INP([PA,A,PS,S],[zEC],
			    DEC(zEC,[PA,A,CA],[],zK,LS2,orig LAs2,
				 NIL))))))

             in	NEW_I([([("i",1)],CA)],
		NEW(CS,
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)])))
             end
    in ("msr3",
	"Improved MSR (n portables $I_i$, one Radio Control Equipment (RCE) $I_{-1}$)",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1)],xKBs),
	 ([("i",1)],zK)	 
	 ])
	   
    end 
(*****************************************************************************)








(*****************************************************************************)
val p4 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,			      
			    INP([PS,S,PA,A,PS,S],[xCSs,xKBs],
                            ADEC(xCSs,[PS,S],[],KUP,LAs4,dest LS4,
			    NEW(Ks,
                            OUT([PA,A,PS,S,AENC([Ks],xKBs,LAs1,dest LS1)],
			    OUT([PA,A,PS,S,ENC([PA,A, AENC([PA,A],KUM,LAs3, pdest LS3)],
					       Ks,LAs2,dest LS2)],
				NIL))))))

                val BProc = OUT([KUP,KSP,A,A,A],NIL)

                val SProc = PAR_I("i",1,
                            OUT([PS,S,PA,A,PS,S,
				 AENC([PS,S],KUM,LS4,pdest LAs4),KSP],
			    INP([PA,A,PS,S],[zEK],
                            ADEC(zEK,[],[zK],KSM,LS1,porig(LAs1),
                            INP([PA,A,PS,S],[zEC],
			    DEC(zEC,[PA,A],[zC],zK,LS2,orig LAs2,
		            ADEC(zC,[PA,A],[],KUP,LS3,dest LAs3,
                            NIL)))))))

             in	ANEW(NAME("K^U",[]),
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)]))
             end
    in ("msr4",
	"Improved MSR (n portables $I_i$, one Radio Control Equipment (RCE) $I_{-1}$). "^
	"Certificates as signatures",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1)],xKBs),
	 ([("i",1)],zK)	 
	 ])
	   
    end 
(*****************************************************************************)


val K    = NAME("K",["i","j"])

val xKB  = VAR("xKB",["i","j"])
val xCS  = VAR("xCS",["i","j"])
val xC   = VAR("xC", ["i","j"])

val LA1   = ("A1",["i"])
val LA2   = ("A2",["i"])
val LA3   = ("A3",["i"])
val LA4   = ("A4",["i"])


val PR    = NAME("J",["j"])
val R     = NAME("R",[])
val CR    = NAME("CR",["j"])

val KRP   = NAMEP("K^R",["j"])
val KRM   = NAMEM("K^R",["j"])

val yEK   = VAR("y1",["i","j"])
val yK    = VAR("y2",["i","j"])
val yEC   = VAR("y3",["i","j"])
val yC    = VAR("y4",["i","j"])


val LR1   = ("R1",["j"])
val LR2   = ("R2",["j"])
val LR3   = ("R3",["j"])
val LR4   = ("R4",["j"])




(*val KSM   = NAMEM("K^S",[])
val CS    = NAME("CS",[])
*)


(*****************************************************************************)
val p5 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,
			    PAR_I("j",1,
			    INP([PR,R,PA,A,PR,R],[xKB],
			    NEW(K,
                            OUT([PA,A,PR,R,AENC([K],xKB,LA1,dest LR1)],
			    OUT([PA,A,PS,S,ENC([PA,A,CA],K,LA2,dest LR2)],
				NIL))))))

                val BProc = PAR_I("j",1,
			    PAR_I("i",1,
                            OUT([PR,R,PA,A,PR,R,KRP],
			    INP([PA,A,PR,R],[yEK],
                            ADEC(yEK,[],[yK],KRM,LR1,porig(LA1),
                            INP([PA,A,PR,R],[yEC],
			    DEC(yEC,[PA,A,CA],[],yK,LR2,orig LA2,
				 NIL)))))))

                val SProc = NIL

             in	NEW_I([([("i",1)],CA)],
		ANEW_I([([("j",1)], NAME("K^R",["j"]))],
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)]))
             end
    in ("msr5",
	"MSR ($n$ portables $I_i$, $n$ Radio Control Equipment (RCE) $I_{-1}$)",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1)],xKBs),
	 ([("i",1)],zK)	 
	 ])
	   
    end 
(*****************************************************************************)

(*****************************************************************************)
val p6 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,
			    PAR_I("j",1,
			    INP([PR,R,PA,A,PR,R],[xKB],
			    NEW(K,
                            OUT([PA,A,PR,R,AENC([K],xKB,LA1,dest LR1)],
			    OUT([PA,A,PR,R,ENC([PA,A,
						AENC([PA,A],KUM,LA3, pdest LR3)],
					       K,LA2,dest LR2)],
				NIL))))))

                val BProc = PAR_I("j",1,
			    PAR_I("i",1,
                            OUT([PR,R,PA,A,PR,R,KRP],
			    INP([PA,A,PR,R],[yEK],
                            ADEC(yEK,[],[yK],KRM,LR1,porig(LA1),
                            INP([PA,A,PR,R],[yEC],
			    DEC(yEC,[PA,A],[yC],yK,LR2,orig LA2,
			    ADEC(yC,[PA,A],[],KUP,LR3,orig LA3,
		            NIL))))))))

                val SProc = OUT([KUP,KUP,KUP,KUP,KUP],NIL)

             in	ANEW_I([([("j",1)], NAME("K^R",["j"]))],
		ANEW(NAME("K^U",[]),		
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)]))
             end
    in ("msr6",
	"MSR ($n$ portables $I_i$, $n$ Radio Control Equipment (RCE) $J_{j}$). "^
	"Certificates are signatures.",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yK)	 
	 ])
	   
    end 
(*****************************************************************************)


(*
(*****************************************************************************)
val p7 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,
			    PAR_I("j",1,
			    INP([PR,R,PA,A,PR,R,CR],[xKB],
			    NEW(K,
                            OUT([PA,A,PR,R,AENC([K],xKB,LA1,dest LR1)],
			    OUT([PA,A,PS,S,ENC([PA,A,CA],K,LA2,dest LR2)],
				NIL))))))

                val BProc = PAR_I("j",1,
			    PAR_I("i",1,
                            OUT([PR,R,PA,A,PR,R,CR,KRP],
			    INP([PA,A,PR,R],[yEK],
                            ADEC(yEK,[],[yK],KRM,LR1,porig(LA1),
                            INP([PA,A,PR,R],[yEC],
			    DEC(yEC,[PA,A,CA],[],yK,LR2,orig LA2,
				 NIL)))))))

                val SProc = NIL

             in	NEW_I([([("i",1)],CA)],
		NEW_I([([("j",1)],CR)],
		ANEW_I([([("j",1)], NAME("K^R",["j"]))],
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)])))
             end
    in ("msr7",
	"IMSR ($n$ portables $I_i$, $n$ Radio Control Equipment (RCE) $I_{-1}$)",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1)],xKBs),
	 ([("i",1)],zK)	 
	 ])
	   
    end 
(*****************************************************************************)
*)



(*****************************************************************************)
val p8 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,
			    PAR_I("j",1,
			    INP([PR,R,PA,A,PR,R],[xKB,xC],
			    ADEC(xC,[PR,R],[],KUP,LA4,orig LR4,
			    NEW(K,
                            OUT([PA,A,PR,R,AENC([K],xKB,LA1,dest LR1)],
			    OUT([PA,A,PR,R,ENC([PA,A,
						AENC([PA,A],KUM,LA3, pdest LR3)],
					       K,LA2,dest LR2)],
				NIL)))))))

                val BProc = PAR_I("j",1,
			    PAR_I("i",1,
                            OUT([PR,R,PA,A,PR,R,KRP,AENC([PR,R],KUM,LR4,pdest LA4),KRP],
			    INP([PA,A,PR,R],[yEK],
                            ADEC(yEK,[],[yK],KRM,LR1,porig(LA1),
                            INP([PA,A,PR,R],[yEC],
			    DEC(yEC,[PA,A],[yC],yK,LR2,orig LA2,
			    ADEC(yC,[PA,A],[],KUP,LR3,orig LA3,
		            NIL))))))))

                val SProc = OUT([KUP,KUP,KUP,KUP,KUP],NIL)

             in	ANEW_I([([("j",1)], NAME("K^R",["j"]))],
		ANEW(NAME("K^U",[]),		
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)]))
             end
    in ("msr8",
	"IMSR ($n$ portables $I_i$, $n$ Radio Control Equipment (RCE) $J_{j}$). "^
	"Certificates are signatures.",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yK)	 
	 ])
	   
    end 
(*****************************************************************************)


end
end