structure Analysis = 
struct
local
open LysaN
open Types
in
(*
   Functions in this file makes it possible to analyse a LySa protocol and retrieve a convenient representation
   of the analysis results.
   
   The results contains the DY knowledge (kappa) and the psi component (authentication). The 
   original analysis result may contain representation of infinit chains of encryptions. These
   encryptions are unfolded to a max depth.
*)



datatype NT = EncNT of NT list * NT * (NT * NT) | NameNT of string | NameT of string (* EncNT is a representation of an encryption possibly containing nonterminals. NameNT is a nonterminal symbol. NameT is a terminal symbol *)

(* From the analysis *)
type psi = (NT * NT) list
type kappa = NT list
type sigma = (NT * NT list * NT * (NT * NT)) list
type c2 = (NT * NT) list



fun isNonterminal(s) = String.isPrefix "L" s  orelse "_L_DY" = s (* all nonterminals start with "L" *)

fun lastChar(str) = String.extract(str,size str - 1,NONE)

fun ntToString(EncNT(nts,key,(cp1,cp2))) =  "{"^ntsToString(nts)^"}"^ntToString(key)
  | ntToString(NameNT(s)) = s
  | ntToString(NameT(s)) = if s = "BULLET" then "I" else if lastChar(s) = "_" then String.substring(s,0,String.size(s)-1) else s

and ntsToString'(nt::nts) = ntToString(nt)^","^ntsToString'(nts)
  | ntsToString' [] = ""

and ntsToString(nts) = let val str = ntsToString'(nts)
                       in String.substring(str,0,String.size(str)-1) end

fun containsNonterminal(EncNT(nts,key,(cp1,cp2))) = List.exists (fn nt => containsNonterminal nt) (key::cp1::cp2::nts)
  | containsNonterminal(NameNT(s)) = true
  | containsNonterminal(NameT(s)) = false

fun containsAssymmetricKeys(EncNT(nts,key,(cp1,cp2))) = List.exists (fn nt => containsAssymmetricKeys nt) (key::cp1::cp2::nts)
  | containsAssymmetricKeys(NameNT(s)) = false
  | containsAssymmetricKeys(NameT(s)) = s = "_MINUS__NDY" orelse s = "_PLUS__NDY"

fun dropLastChar(s) = String.substring(s,0,String.size(s)-1)

(* K_n => KEY[n], _NDY => I, BULLET => I *)
fun toNT(s) = if isNonterminal(s) then NameNT(s) else NameT(s)
(*                if String.isPrefix "K'" s andalso s <> "K'" then NameT("KEY["^dropLastChar(String.substring(s,2,String.size(s)-2))^"]") else
                if String.substring(s,String.size(s)-1,1) = "_" orelse String.substring(s,String.size(s)-1,1) = "'" then NameT(String.substring(s,0,String.size(s)-1)) else
                case s of 
                  "_NDY" => NameT("I")
                 |"BULLET" => NameT("I")
                 | s => NameT(s)
              
*)
fun analyze(pLysaN,principals) = let val pLysa = LysaN2Lysa.convertS(pLysaN,principals)
                                     val pL = Lysa2LLysa.l2llProc(pLysa)
                                     val ALFP = Analysis1.analyse(pL)^
                                                "& (A x. A y. N(y) & C(x,y) => OBS(x))"^
                                                "& (A L0S. A ELL. A LS. A ELL0. PQD(L0S, ELL, LS, ELL0) => "^
                                                "((!D(L0S, ELL) | !D(LS, ELL0)) => PSIFull(ELL0, ELL,LOS,LS)))"
                                     val ss   = FormulaAnalyzer.init()
                                 in (FormulaAnalyzer.solveStd(ss,FormulaAnalyzer.inputStr(ALFP));FormulaAnalyzer.solveData(ss,FormulaAnalyzer.inputStr(ALFP))) end

fun analyzeLysa(pLysa) = let val pL = Lysa2LLysa.l2llProc(MLysa2Lysa.ml2lProc 1 pLysa)
                                        val ALFP = Analysis1.analyse(pL)^
                                                "& (A x. A y. N(y) & C(x,y) => OBS(x))"^
                                                "& (A L0S. A ELL. A LS. A ELL0. PQD(L0S, ELL, LS, ELL0) => "^
                                                "((!D(L0S, ELL) | !D(LS, ELL0)) => PSIFull(ELL0, ELL,LOS,LS)))"
                                     val ss   = FormulaAnalyzer.init()
                   (*                  val _ = FormulaAnalyzer.solveStd(ss,FormulaAnalyzer.inputStr(ALFP))*)
                                 in FormulaAnalyzer.solveData(ss,FormulaAnalyzer.inputStr(ALFP)) end

fun analyzeLysaPrint(pLysa) = let val pL = Lysa2LLysa.l2llProc(MLysa2Lysa.ml2lProc 1 pLysa)
                                        val ALFP = Analysis1.analyse(pL)^
                                                "& (A x. A y. N(y) & C(x,y) => OBS(x))"^
                                                "& (A L0S. A ELL. A LS. A ELL0. PQD(L0S, ELL, LS, ELL0) => "^
                                                "((!D(L0S, ELL) | !D(LS, ELL0)) => PSIFull(ELL0, ELL,LOS,LS)))"
                                     val ss   = FormulaAnalyzer.init()
                                 in FormulaAnalyzer.solveStd(ss,FormulaAnalyzer.inputStr(ALFP)) end



fun getRelation(name,(univers,(name',ls)::relList)) = if name = name' then SOME ls else getRelation(name,(univers,relList))
  | getRelation(_,(_,[])) = NONE


fun getKappa'(21,_)       = []
  | getKappa'(i,solution) = case getRelation("KAPPA"^Int.toString(i)^"/"^Int.toString(i),solution) of
                             SOME(ls) => ls @ getKappa'(i+1,solution)
                            |NONE => getKappa'(i+1,solution)


fun getKappa(solution) = List.map toNT (Aux.removeDoubles(List.foldr (fn (ls1,ls2) => ls1@ls2) [] (getKappa'(1,solution))))
       
fun getPsi'(solution) = case getRelation("PSI/2",solution) of NONE => [] | SOME rel => rel

fun getPsi(solution) = List.map (fn [x1,x2] => (toNT x1,toNT x2) | _ => raise Problem("")) (getPsi'(solution))


(* parses a representation of an encryption of an i tuple [l1,l2,l3,...,l(3+i-1),l(3+i),l(3+i+1)] => l1 = ENC{l3,...,l(3+i-1)}l2(l(3+i),l(3+i+1))*)
fun convertSigma i (nt::key::sigmai) = let val ls = List.take(sigmai,i)
                                       in case List.drop(sigmai,i) of [cp1,cp2] => (nt,ls,key,(cp1,cp2)) end

fun getSigma'(24,solution) = []
  | getSigma'(i,solution) = case getRelation("SIGMA"^Int.toString(i)^"/"^Int.toString(i+4),solution) of
                             NONE => getSigma'(i+1,solution)
                            |SOME(sigmai) => (List.map (convertSigma i) sigmai)@getSigma'(i+1,solution)

fun getSigma(solution) = List.map (fn (nt,nts,key,(cp1,cp2)) => (toNT nt,EncNT(List.map toNT nts,toNT key,(toNT cp1,toNT cp2)))) (getSigma'(1,solution))


fun searchSigma solution (NameNT(s)) = let val rel = getSigma(solution)
                                           val rel = List.filter (fn (nt1,enc) => nt1 = NameNT(s)) rel
                                       in List.map (fn (nt1,enc) => enc) rel end
  | searchSigma solution _ = []

fun getC2(solution) = case getRelation("C/2",solution) of NONE => raise Problem("") | SOME(rel) =>
                        List.filter (fn (x1,x2) => x1 <> x2) (List.map (fn [x1,x2] => (toNT x1,toNT x2) | _ => raise Problem("")) rel)

fun searchC2 solution (NameNT(s)) = let val rel = getC2(solution)
                                        val rel = List.filter (fn (nt1,nt2) => nt1 = NameNT(s)) rel
                                    in List.map (fn (nt1,nt2) => nt2) rel end
  | searchC2 solution _ = []

fun getD2(solution) = case getRelation("D/2",solution) of NONE => raise Problem("") | SOME(rel) =>
                        List.filter (fn (x1,x2) => x1 <> x2) (List.map (fn [x1,x2] => (toNT x1,toNT x2) | _ => raise Problem("")) rel)

fun searchD2 solution (NameNT(s)) = let val rel = getD2(solution)
                                        val rel = List.filter (fn (nt1,nt2) => nt1 = NameNT(s)) rel
                                    in List.map (fn (nt1,nt2) => nt2) rel end
  | searchD2 solution _ = []
                                   

(* returns a list of encryptions based on all combinations of one unfolding of all nonterminals in the input encryption *)
fun unfoldEncryption solution i (EncNT(nts,key,(cpnt1,cpnt2))) = let val ntss = List.map (unfold solution i) nts
                                                                     val keys = unfold solution i key
     	                                                             val cpnt1s = unfold solution i cpnt1
                                                                     val cpnt2s = unfold solution i cpnt2
                                                                     val perms = Perm.enumeratePermutations2(keys::cpnt1s::cpnt2s::ntss)
                                                                in List.map (fn key'::cpnt1'::cpnt2'::nts' => EncNT(nts',key',(cpnt1',cpnt2'))) perms end
  | unfoldEncryption solution i _ = raise (Problem("not an encryption"))

(* returns a list of possible values corresponding to the input symbol. Unfolded to encryption level i *)
and unfold solution i (NameT(s))  = [NameT(s)]
  | unfold solution i (NameNT(s)) = let val rel = (searchC2 solution (NameNT(s)))@(searchD2 solution (NameNT(s)))
                                        val rel = if i>0 then rel@(searchSigma solution (NameNT(s))) else rel
                                        val rel = if rel = [] then [NameNT(s)] else rel
                                    in if i > 0 then Aux.removeDoubles(Aux.meltList(List.map (unfold solution i) rel)) else rel end
  | unfold solution i enc         = let val unf = unfoldEncryption solution (i-1) (enc)
 (*                                     val unf = if unf = [] then [enc] else unf *)
                                    in if i > 0 then Aux.removeDoubles(Aux.meltList(List.map (unfold solution (i-1)) unf)) else unf end

fun masterunfold(solution,i,nt) = List.filter (fn nt => not (containsNonterminal nt)) (unfold solution i nt)


fun unfoldKappa(solution,i) = let val kappa = getKappa(solution)
                            in Aux.removeDoubles(Aux.meltList(List.map (unfold solution i) kappa)) end

fun unfoldLDY(solution,i) = let val ls = List.filter (fn nt => not (containsAssymmetricKeys nt)) (unfold solution i (NameNT"_L_DY"))
                                val ls = Aux.removeDoubles(ls)
                                val (l1,l2) = List.partition (fn nt => not (containsNonterminal nt)) ls
                            in l1(*@l2*) end

fun solutionToString(solution) = let val c2 = (List.foldl (fn ((nt1,nt2),str) => "("^ntToString(nt1)^","^ntToString(nt2)^"),"^str) "" (getC2(solution)))
                                     val sigma = (List.foldl (fn ((nt1,nt2),str) => "("^ntToString(nt1)^","^ntToString(nt2)^"),"^str) "" (getSigma(solution)))
                                     val kappa = (List.foldl (fn (nt,str) => ntToString(nt)^","^str) "" (getKappa(solution)))
                                     val psi = (List.foldl (fn ((nt1,nt2),str) => "("^ntToString(nt1)^","^ntToString(nt2)^"),"^str) "" (getPsi(solution)))
                                 in 
                                   "c2 =\n"^c2^"\n"^
                                   "sigma =\n"^sigma^"\n"^
                                   "kappa =\n"^kappa^"\n"^ 
                                   "psi =\n"^psi^"\n"
                                 end
fun psiToString(solution) =   let val psi = (List.foldl (fn ((nt1,nt2),str) => "("^ntToString(nt1)^","^ntToString(nt2)^"),"^str) "" (getPsi(solution)))
                              in  "psi =\n"^psi^"\n" end
                              
end
end