structure Constraints =
struct

exception Search of string

local
open Types
open Aux
open Graph
open LysaN
open ToString

(*
Constraint generation. Creating the foundation for solving the constraints.

Constraints are generated from a given path by going through the input- and output-steps one at a time.
An output-step gives rise to an addition to the knowledge of the intruder. An input-step gives rise to
a constraint.
A constraint is constructed by taking a term from the input and pairing it with the current IK.
All terms encountered are paired with an index. An index corresponding to a given term is increased if:
The term is a variable and the defining occurence of the variable is encountered
The term is a name and the "new construction" of the name is encountered
This ensures that we can tell different instances of a name appart, and that different instances of the 
same variable can be assigned different values. The problem occurs when a path contains a cycle.
These index are not part of the analysis result. This means that we cannot know how to specify the right
index when specifying the breach we want to search for. This gives rise to the notion of a 
\emph{wild card index}, 0. Whenever we compare two terms thay are only equal when their index are equal. 
All index are equal to the wild card index 0.
All terms in the intruder knowlege are furthemore decorated with a boolean value stating whether this term
has been considered with regard to decryption. I a term has been considered it means that we are in a branch
of the search and the term has either been decrypted in this branch or another branch and the decryption of
the term should not be considered again. All terms has to be considered before moving on to a new state in
the constraint solving.
Terms that need indexing are called instance terms. It is names that occur somewhere in the looping process
in the construction NEW(NAME n). And it is all variables. All other terms are given the wildcard indez.
Since the first occurence instance terms are the defining occurence of a variable and the new construction of
a name respectively, it is easy to keep track of when to increase the index.
*)

in

type constraint = CTerm * CTerm list

infix ==

(* a member function which ignores the flags of the terms in IK *)
fun memberOfIK(t,IK) = List.exists (fn t' => t == t') IK

fun createConstraints(t::ts,IK,termList) = let val (t',termList') = termToCTerm(t,termList) 
                                               val (constraints,termList'') = createConstraints(ts,IK,termList')
                                           in ((t',IK)::constraints,termList'') end
  | createConstraints([],IK,termList) = ([],termList)


(* union function which adds ts to IK without any doubles *)                                                                                                                      
fun addToIK(ts,IK) = let val (ts',termList') = termToCTermList(ts,[]) (* convert to CTerm *)
                         val ts'' = (List.filter (fn t => not(memberOfIK(t,IK))) ts') (* Filter doubles *)
                     in ts''@IK end

val IK = [CNAME("I",0),CKEY(CNAME("I",0)),CNAME("A",0),CNAME("B",0),CNAME("S",0)]
val dummyVar = CVAR("z",0)

(* takes a path, an initial IK and an optional breach term and creates the constraints *)                                                                                                                      
fun createCS(nodeRef1::nodeRef2::path,IK) = (
  case (lookUpTransition(nodeRef1,nodeRef2)) of
     (_,OUTSTEP(terms)) => createCS(nodeRef2::path,addToIK(terms,IK))
    |(_,INSTEP(terms)) => (#1 (createConstraints(terms,IK,[])))@createCS(nodeRef2::path,IK)
  )
  | createCS([nodeRef],IK) = [(dummyVar,IK)] (* If we have a secrecy breach it is added as a constant constraint in the end of a constriantset*)
  | createCS([],IK) = [(dummyVar,IK)]

fun createConstraintSet(path) = createCS(path,IK)

 

fun initConstraint(leakterms) = let val (leakterms',_)= termToCTermList(leakterms,[]) in (dummyVar,IK@leakterms') end



fun psiToString(((t1,t2),(t1',t2'))::psi) = "(("^cTermToString(t1)^","^cTermToString(t2)^")("^cTermToString(t1')^","^cTermToString(t2')^"))\n"^psiToString(psi)
  | psiToString [] = ""


(* constraints to string *)
fun constraintToString(t,IK) = cTermToString(t)^":"^cTermListToString(IK)
fun constraintSetToString(c::cs) = constraintToString(c)^"\n"^constraintSetToString(cs)
  | constraintSetToString([]) = ""


end
end                               