structure DY =
struct

local
open Types
open LysaN
in
(* functions for testing whether a term with no free variables is element in a set of terms under the generation and analysis rules of DY attacker *)
(* An encryption term t can be element of a DY-knowledge even though the crypto points does not match. The mismatch is recorden. *)

(* {%x}k(A<-B) {M}K(B'->A')                                           *)
(* if A <> A' or B <> B' then SOME((A,A'),(B,B')) else NONE               *)

type mismatch = ((Term*Term)*(Term*Term)) list

(* 
  how to read the mismatch:
    ((cp1,cp1'),(cp2,cp2')) corresponds to a mismatch between E and E' with crypto points 
    (cp1,ORIG(cp2)) and (cp2',DEST(cp1')).
    The mismatch occurs when either cp1 <> cp1' or cp2 <> cp2'.
    In other words: 
    1. We are cp1 and the encryption was intended for cp1'.
    2. We are expecting a message from cp2, but the message comes from cp2'.
    In other words:
    ((actual reciever,intended reciever),(intended sender,actual sender))
    
    (orig,dest) corresponds to (actual reciever,actual sender). No information abour intended.
    
    From the analysis we have the set psi of pairs (orig,dest).
    This is a simple version of the mismatch used here. Orig and dest are the actual origin and destination
    of an encryption. It does not contain information about intended origin and destination.
    Between ((cp1,cp1'),(cp2,cp2')) and (orig,dest) we have the correspondance, orig = cp2', dest = cp1.
*)

(* if the crypto points lead to a mismatch it is added to the list mismatches *)
fun addMismatch((cp1,ORIG(cp2)),(cp2',DEST(cp1')),mismatches) = if cp1 <> cp1' orelse cp2 <> cp2' 
                                                                then ((cp1,cp1'),(cp2,cp2'))::mismatches 
                                                                else mismatches
  | addMismatch _ = raise Problem("Orig Dest")


fun equalIgnoreCpsList(t::ts,t'::ts') = let val (b1,mismatch1) = equalIgnoreCps(t,t')
                                            val (b2,mismatch2) = if b1 then equalIgnoreCpsList(ts,ts') else (false,[])
                                        in (b1 andalso b2,mismatch1 @ mismatch2) end 
  | equalIgnoreCpsList([],[]) = (true,[])
  | equalIgnoreCpsList _ = (false,[])

and equalIgnoreCps(NAME(n),NAME(n')) = (n = n',[])
  | equalIgnoreCps(VAR(x),VAR(x')) = (x = x',[])
  | equalIgnoreCps(KEY(t),KEY(t')) = equalIgnoreCps(t,t')
  | equalIgnoreCps(ENC(ts,key,cps),ENC(ts',key',cps')) = 
                  let val (b1,mismatch1) = equalIgnoreCpsList(ts,ts') 
                      val (b2,mismatch2) = if b1 then equalIgnoreCps(key,key') else (false,[])
                  in (b1 andalso b2,addMismatch(cps,cps',mismatch1@mismatch2)) end  (* the only place where we generate a new mismatch *)
  | equalIgnoreCps _ = (false,[])

fun memberIgnoreCps(t,t'::ts) = let val (b1,mismatch1) = equalIgnoreCps(t,t') 
                                    val (b2,mismatch2) = memberIgnoreCps(t,ts)
                                in (b1 orelse b2,(if b1 then mismatch1 else [])@(if b2 then mismatch2 else [])) end
  | memberIgnoreCps(t,[]) = (false,[])


(* searches recursively in deeper levels of encryption among terms in ts *)
fun cryptMember'(t,ENC(ts',_,_)::ts) = cryptMember'(t,ts') orelse cryptMember'(t,ts)
  | cryptMember'(t,t'::ts) = t = t' orelse cryptMember'(t,ts)
  | cryptMember'(t,[]) = false




(* returns the encryption which contains the term t or a subpattern containing t*)
fun cryptMember(t,ENC(tsEnc,k,cps)::ts) = if #1 (memberIgnoreCps(t,tsEnc))       (* does t exist on this level of current encryption? *)
                                          orelse cryptMember'(t,tsEnc) (* or any sublevels *)
                                          then SOME(ENC(tsEnc,k,cps))  (* then return the encryption *) 
                                          else cryptMember(t,ts)       (* else search the rest of the list *)
  | cryptMember(t,t'::ts) = cryptMember(t,ts)
  | cryptMember(t,[]) = NONE

(*Tests that all elements of a list are elements in another list*)
fun elementInTs(t::ts,ts') = let val (b1,mismatch1) = elementIn'(t,ts') 
                                 val (b2,mismatch2) = if b1 then elementInTs(ts,ts') else (false,[])
                             in (b1 andalso b2,mismatch1@mismatch2) end
  | elementInTs([],_) = (true,[]) 

and canBeComposed(ENC(ts',key,cps as (CP,ORIG(CPorig))),ts) = 
                                   let val (b1,mismatch1) = elementInTs(ts',ts)
                                       val (b2,mismatch2) = elementIn'(key,ts)
                                     val mismatchDY = ((CP,CP),(CPorig,NAME("l_bullet")))  (* a mismatch is created since DY can create the encytionterm *)
                                   in  (b1 andalso b2,mismatchDY::mismatch1@mismatch2) end
  | canBeComposed _ = raise Problem("not an encryption")

and nonEncElementIn(t,ts) = let val b1 = Aux.member(t,ts) (*is the term on first level of ts?*)
                                val (b2,mismatch2) = 
                                  case cryptMember(t,ts) of (* search for the term t in encryptions of any level among terms, ts. Returns the encryption on current level with a subpattern containing t. The encryption is attempted decrypted. If successfull, the encryption is removed and the terms of the encryption is added to the knowledge and the search is restarted *)
                                    NONE => (false,[])
                                   |SOME(enc as ENC(tsEnc,key,cps)) => 
                                      if #1 (elementIn'(key,ts)) (* no need to save mismatch information at this point since we are decrypting and restarting the search *)
                                      then let val ts' = Aux.remove(enc,ts)@tsEnc in elementIn'(t,ts') end
                                      else (false,[])
                                  
                            in (b1 orelse b2,if b2 then mismatch2 else []) end

and elementIn'(t,ts) = let val (b1,mismatch1) = memberIgnoreCps(t,ts) (* is t member on this level ?*)
                           val (b2,mismatch2) = (* does t exist in an encryption on some level and can we unfold one step towards this encryption *)
                                  case cryptMember(t,ts) of (* search for the term t in encryptions of any level among terms, ts. Returns the encryption on current level with a subpattern containing t. The encryption is attempted decrypted. If successfull, the encryption is removed and the terms of the encryption is added to the knowledge and the search is restarted *)
                                    NONE => (false,[])
                                   |SOME(enc as ENC(tsEnc,key,cps)) => 
                                      if #1 (elementIn'(key,ts)) (* no need to save mismatch information at this point since we are decrypting and restarting the search *)
                                      then let val ts' = Aux.remove(enc,ts)@tsEnc in elementIn'(t,ts') end
                                      else (false,[])
                           val (b3,mismatch3) = case t of ENC(tsEnc,key,cps) => canBeComposed(t,ts) | _ => (false,[])
                       in (b1 orelse b2 orelse b3,(if b1 then mismatch1 else [])@(if b2 then mismatch2 else [])@(if b3 then mismatch3 else [])) end
                          

fun elementIn(t,ts) = let val (b,mismatch) = elementIn'(t,ts) in (b,Aux.removeDoubles(mismatch)) end




(*strategy:
Search for t in DY(ts) 


1. is t member of ts on level 0?
2. does t exist in an encryption on some level? If so, unfold one level of that branch of encryption and restart the search.
3. If t is an encryption, can t be composed? This is determined by calling elementIn' on each of the parts of the encryption, t.

All three results are merged. The final list of mismatches should be cleaned for doubles.



More complicated strategy:

Two cases:
1. t is a simple term (not encryption term). nonEncElementIn()
2. t is an encryption term.                  encElementIn() 

case 1.
We check two things:
  1. We check whether t is in ts. Result: b1
  2. We search inside encryptions of ts in arbitrary deep levels.
     If we find t in encryption e or a sublevel of e, we check whether the key of e exists in DY(ts / e).
     If this is the case, we remove e from ts and add the content of e to ts. I.e. we decrypt e. Then we repeat the search.
       result: (b2,mismatch2)
The results from 1 and 2 are merged. (b1 orelse b2,if b2 then mismatch2 else [])

case 2.
We have to follow the steps above as well as check if we can find each of the terms 
which constitutes the encryption t in ts.
If we can find t in ts or sublevels, then the match might give rise to a mismatch of cps.
If we can construct t from ts, then we have a mismatch of type (A<-B) (l*->_) => (A,B,l*,A)
We have to check both cases, because even though we can use one of them, we also need to collect the 
mismatch from the other.
*)
end
end