structure Graph =
struct

exception Search of string
local
open Types
open Aux
open ToString
open LysaN
in
datatype node = NODE of {s:state, transOut:(transition * node ref) list,transIn:(transition * node ref) list,visits:int}
type graph = node ref list
type path = node ref list

(*
create a new node with an empty transision list
mkNode: state -> node ref
*)
fun mkNode(state) = ref(NODE{s=state,transOut = [ ],transIn = [], visits = 0})

(*
returns the state of a node ref
*)
fun getState(nodeRef) = let val NODE(n) = !nodeRef
                        in #s n end

(*
returns the transitions of a node
*)
fun getTransOut(nodeRef) = let val NODE(n) = !nodeRef
                           in (#transOut n) end

(*
returns the transitions of a node
*)
fun getTransIn(nodeRef) = let val NODE(n) = !nodeRef
                           in (#transIn n) end

(*
*)
fun getVisits(nodeRef) = let val NODE(n) = !nodeRef
                           in (#visits n) end

(*
returns the children of a node
node ref -> node ref list
*)
fun getChildren(nodeRef) = List.map (fn (tt,nodeRef') => nodeRef') (getTransOut(nodeRef))

(*--------------------------------*)



fun addTransition'(nr1,nr2,t) = (
                                  nr1 := NODE{
                                    s = getState(nr1),
                                    transOut=(t,nr2)::getTransOut(nr1),
                                    transIn = getTransIn nr1,
                                    visits = getVisits nr1
                                  };
                                  nr2 := NODE{
                                    s = getState nr2,
                                    transOut = getTransOut nr2,
                                    transIn=(t,nr1)::getTransIn nr2,
                                    visits = getVisits nr2
                                  }
                                )

(*
finds a node with a given state in a list of nodes. If the node does not exist, it is created.
state * (node ref) list -> node ref * (node ref) list
*)

fun lookUpNodeRef(state,node::nodeRefs) = 
      if state = getState node
      then (node,node::nodeRefs) 
      else let val (node',nodeRefs') = lookUpNodeRef(state,nodeRefs)
           in (node',node::nodeRefs') end
  | lookUpNodeRef(state,[]) = let val nRef = mkNode(state)
                              in(nRef,[nRef]) end

(*
takes two states, a transtype and a list of nodes. A transition is added between the 
nodes representing the two states. If nodes does not exist yet they are created. The list
Nodes contains existing nodes and are updated if new nodes are created.
state*state*transition*node ref list -> node ref list
*)
fun addTransition(s1,s2,transition,graph) = 
    let val (n1,graph) = lookUpNodeRef(s1,graph)
        val (n2,graph) = lookUpNodeRef(s2,graph)
    in (addTransition'(n1,n2,transition);(n2,graph)) end






(*
a simple breadth first search using a queue for visited nodes.
f has the type t -> unit
*)
fun simpleBFS f (nodeRef::queue,visited) = if member(nodeRef,visited)
                                           then simpleBFS f (queue,visited)
                                           else let val children = getChildren(nodeRef)
                                                in (f(nodeRef);simpleBFS f (queue@children,nodeRef::visited)) end
  | simpleBFS f ([],_) = ()


fun bfsFold f e (nodeRef::queue,visited) = if member(nodeRef,visited)
                                           then bfsFold f e (queue,visited)
                                           else let val children = getChildren(nodeRef)
                                                in f(nodeRef,bfsFold f e (queue@children,nodeRef::visited)) end
  | bfsFold f e ([],_) = e


(*
toString for a path (list of node ref)
*)
fun pathToString(nodeRef::queue) = "("^stateToString'(getState(nodeRef))^") "^pathToString(queue)
  | pathToString([]) = ""

fun pathsToString(p::paths) = pathToString(p)^"\n"^pathsToString(paths)
  | pathsToString [] = ""


(*
BFS without queue. Searches each state muiltple times. max path length = p.
*)
fun dfs(0,nodeRef,path) = print(pathToString(nodeRef::path)^"\n")
  | dfs(p,nodeRef,path) = (*(print("calling dfs' with "^pathToString(getChildren(nodeRef))^", path is: "^pathToString(path)^"\n");*)
                          dfs'(p,getChildren(nodeRef),nodeRef::path)

and dfs'(p,nodeRef::children,path) =  (dfs(p-1,nodeRef,path);dfs'(p,children,path))
  | dfs'(_,[],_) = ()

(*
uses simpleBFS to print each node
*)                                                                                                                    
fun bfPrint(nodeRef) = simpleBFS (fn nodeRef => let val NODE(n) = !nodeRef                                            
                                                in print(stateToString(#s n)^"\n") end                                
                                 ) ([nodeRef],[])                                                                     

fun lookUpTransition(nodeRef1,nodeRef2) =  let val trans = getTransOut(nodeRef1)
                                               val find = List.find (fn (t,nodeRef) => nodeRef2 = nodeRef) trans
                                           in case find of
                                              NONE => (print ("!!("^stateToString'(getState(nodeRef1))^")-/-> ("^stateToString'(getState(nodeRef2))^")!!\n");raise Problem ("no transition between nodeRefs in lookUpTransition:"))
                                             |SOME(transition,_) => transition
                                           end

datatype cStep = COUTSTEP of string * Types.CTerm list | CINSTEP of string * CTerm list


fun createNarration(n1::n2::path,namelist,sub) = 
  let val (principal,comstep) = lookUpTransition(n1,n2)
  in (
    case comstep of
      OUTSTEP(ts) => let val (cts,namelist') = termToCTermList(ts,namelist)
                     in COUTSTEP(principal,List.map (applySubstitution sub) cts)::createNarration(n2::path,namelist',sub) end
     |INSTEP(ts) =>  let val (cts,namelist') = termToCTermList(ts,namelist)
                     in  CINSTEP(principal,List.map (applySubstitution sub) cts)::createNarration(n2::path,namelist',sub) end
  )
  end
  | createNarration(_) = []


(*tex
fun narrationToString(cStep::steps) = 
    (
      case cStep of
      COUTSTEP(principal,ts) => "&"^principal^"& \\rightarrow &I&:&\\OUTS{"^cTermListToString(ts)^"}"
     |CINSTEP(principal,ts) => "&I& \\rightarrow &"^principal^"&:&\\INPS{"^cTermListToString(ts)^"}"
    )^"\n"^narrationToString(steps)
  | narrationToString [] = "\n"
*)
(*without tex*)
fun narrationToString(cStep::steps) = 
    (
      case cStep of
      COUTSTEP(principal,ts) => principal^" -> I: <"^cTermListToString(ts)^">"
     |CINSTEP(principal,ts) => "I -> "^principal^":("^cTermListToString(ts)^")"
    )^"\n"^narrationToString(steps)
  | narrationToString [] = "\n"


(*                                                                                                                    
                                                                                                                      
*)                                                                                                                    
fun transitionToString(principal,OUTSTEP(ts)) = principal^":<"^termListToString(ts)^">"                                                    
  | transitionToString(principal,INSTEP(ts))  = principal^":("^termListToString(ts)^")"                                                     

(*                                                                                                                    
                                                                                                                      
*)                                                                                                                    
fun transitionsToString'(state,(trans,nodeRef)::nodeRefs) =                                                           
        stateToString(state)^"->"^stateToString(getState(nodeRef))^                                                   
        "[label = "^quotes(transitionToString(trans))^"]"^                                                            
        "\n"^                                                                                                         
        transitionsToString'(state,nodeRefs)                                                                          
  | transitionsToString'(_,[]) = ""                                                                                     
                                                                                                                      
(*                                                                                                                    
                                                                                                                        
*)                                                                                                                      
fun transitionsToString(nodeRef) = transitionsToString'(getState(nodeRef),getTransOut(nodeRef))                           

                                                                                                                    
fun listTransitions(nodeRef) = List.map (fn (t,nodeRef) => getState(nodeRef)) (getTransOut(nodeRef))                  
                                                                                                                          
fun graphToDot(nodeRef) =  bfsFold (fn (nodeRef,s) => s^transitionsToString(nodeRef)) "" ([nodeRef],[])

fun graphToDotFile(nodeRef,filename) =  
      let val g = "digraph wmf{\n"^
                   graphToDot(nodeRef)^
                   "}"
          val file = TextIO.openOut("graph/"^filename^".dot")  
          val _ = TextIO.output(file, g)
          val _ = TextIO.closeOut(file)
(*      in OS.Process.system("dot -Tps2 -ograph/"^filename^".ps graph/"^filename^".dot") end*)
      in OS.Process.system("dot -Tgif -ograph/"^filename^".gif graph/"^filename^".dot") end


fun searchFor name (NODE(n)) = #s n = name

fun stateListToPath(s::states,graph) = let val (nodeRef,graph) = lookUpNodeRef(s,graph)
                                       in nodeRef::stateListToPath(states,graph) end
  | stateListToPath([],_) = []

fun extendPath(path) = let val pathRev = rev path
                       in List.map (fn node => rev(node::pathRev)) (getChildren(hd(pathRev))) end
                       
fun extendPaths(paths) = Aux.meltList(List.map extendPath paths)


end
end                               