(* These functions makes it possible to convert from LysaN to Lysa. The only part of the conversion wich is not straight foreward is
   the convertion of input with patterns containing free variables. Two problems must be addressed:
     1. The ordering of free variables on the form: (E1,...Ei;xj,...,xk)
     2. Conversion of an encryption pattern which contains free variables (possibly nested in several layers of encryption). Example:
       (E(k)(%x)).P -> (;xGen).decrypt xGen as E(k)(%x).P

*)

structure LysaN2Lysa =
struct
open LysaN
open Types
datatype PROC' = INPUT' of Term list | DECRYPT' of Term * Term list * Term * (Term * Form)

val NDY = Lysa.NAME("_NDY",[])

val index = ref 0;
fun getVar() = (index := !index + 1 ; "x"^Int.toString(!index))
fun resetVar() = index := 0

fun containsDefVars(DEFVAR(_)) = true
  | containsDefVars(ENC(ts,_,_)) = List.exists containsDefVars ts
  | containsDefVars _ = false

(*flatten a list of terms. The list could originate from an encryption or an input 
  an encryption-term is transformed if it contains free variables.*)
fun flatten(TS as ENC(ts',k,(cp,form))::ts,ps) = let val (tsRest,psRest) = flatten(ts,ps)
                                                 in
                                                   if true (*containsDefVars(hd TS)*)
                                                   then let val x = getVar()
                                                        in (DEFVAR(x)::tsRest,DECRYPT'(VAR(x),ts',k,(cp,form))::psRest) end
                                                   else ((hd TS)::tsRest,psRest)
                                                 end
  | flatten(t::ts,ps)                  = let val (tsRest,psRest) = flatten(ts,ps) in (t::tsRest,psRest) end
  | flatten([],ps)                     = ([],ps)

(*flatten an input or decrypt*)
fun flattenP(INPUT'(ts))         = let val (ts',ps) = flatten(ts,[]) in INPUT'(ts')::ps end
  | flattenP(DECRYPT'(t,ts,key,(cp,form))) = let val (ts',ps) = flatten(ts,[]) in DECRYPT'(t,ts',key,(cp,form))::ps end

(*flatten a list of inputs and decrypts*)
fun flattenPList(p::ps) = let val ps' = flattenP(p) in hd(ps')::flattenPList(tl(ps')@ps) end
  | flattenPList [] = []

fun convertSymbol(s) = (s,[])

fun convertCP(NAME(s)) = if s = "I" then (Lysa.CPDY) else (s,[])
  | convertCP _ = raise Problem("")

fun convertForm(ORIG(NAME n)) = SOME(Lysa.ORIG([convertCP(NAME n)]))
  | convertForm(DEST(NAME n)) = SOME(Lysa.DEST([convertCP(NAME n)]))
  | convertForm(_) = raise Problem("")

fun convertT(VAR(x)) = Lysa.VAR(convertSymbol(x))
  | convertT(NAME n) = Lysa.NAME(n,[])
  | convertT(KEY n) = (case n of 
                          NAME(n) => Lysa.NAME("K'"^n,[])
                        | t => (print("**"^ToString.termToString(t)^"**\n");raise Problem(""))
                      )
  | convertT(ENC(ts,t,(cp,form))) = Lysa.ENC(convertTS(ts),convertT(t),convertCP(cp),convertForm(form))
  | convertT(_) = raise Problem("All other terms should have been replaced at this point")

and convertTS(ts) = List.map convertT ts

fun split(t::ts) = let val (ts,vs) = split(ts)
                   in case t of
                      DEFVAR(x) => (ts,Lysa.VAR(convertSymbol x)::vs)
                     |t         => (convertT(t)::ts,vs)
                   end
  | split [] = ([],[])

fun convertPList(INPUT'(ts)::ps,P)         = let val (ts,vs) = split(ts) in Lysa.INP(ts,vs,convertPList(ps,P)) end
  | convertPList(DECRYPT'(t,ts,key,(cp,form))::ps,P) = 
      let val (ts,vs) = split(ts) 
      in Lysa.DEC(convertT(t),ts,vs,convertT(key),convertCP(cp),convertForm(form),convertPList(ps,P)) end
  | convertPList([],P) = P
  
fun convertP(OUT(ts,P)) = Lysa.OUT(convertTS(ts),convertP(P))
  | convertP(INP(ts,P)) = let val ps = flattenPList [INPUT' ts] in convertPList(ps,convertP(P)) end
  | convertP(NEW(n,P))  = Lysa.NEW(convertT(n),convertP(P))
  | convertP(NIL) = Lysa.NIL

(*********  expand   **********)
fun pow(r,p) = Real.round(Math.pow(Real.fromInt(r),Real.fromInt(p)))

fun convert(n,r,0) = [n]
  | convert(n,r,p) = let val d = pow(r,p)
                         val c = n div d
                     in if c = 0 then 0::convert(n,r,p-1) else c::convert(n mod d,r,p-1) end

fun intListToPrincipals(n::ns,ps) = List.nth(ps,n)::intListToPrincipals(ns,ps)
  | intListToPrincipals([],_) = []

fun enumeratePermutations'(~1,_,_,_) = []
  | enumeratePermutations'(i,r,p,principals) = intListToPrincipals(convert(i,r,p),principals)::enumeratePermutations'(i-1,r,p,principals)

fun enumeratePermutations(principals,n) = let val l = List.length(principals)
                                              val max = pow(l,n)-1
                                          in enumeratePermutations'(max,l,n-1,principals) end

fun defvarpTs [] = []
  | defvarpTs(ENC(ts,t,_)::ts') = defvarpTs(ts)@defvarpTs(ts')
  | defvarpTs(t::ts) = case t of DEFVARP(s) => t::defvarpTs(ts) | _ => defvarpTs(ts)

fun defvarp(INP(ts,P)) = defvarpTs(ts)@defvarp(P)
  | defvarp(OUT(ts,P)) = defvarpTs(ts)@defvarp(P)
  | defvarp(NEW(n,P))  = defvarp(P)
  | defvarp NIL = []

fun mapsTo(t,(v,p)::vps) = (case (t,v) of 
                             (DEFVARP(v1),DEFVARP(v2)) => if v1 = v2 then SOME p else mapsTo(t,vps)
                            |(VAR(v1),DEFVARP(v2))     => if v1 = v2 then SOME p else mapsTo(t,vps)
                            | _                        => mapsTo(t,vps))
  | mapsTo(_,[])         = NONE

(* The following functions replaces varps and adds corresponding indices to variables and fresh names *)
fun replaceVarpsCP(index,sub,nnscope,(cp,ORIG(cp'))) = (replaceVarpsT(index,sub,nnscope,cp),ORIG(replaceVarpsT(index,sub,nnscope,cp')))
  | replaceVarpsCP(index,sub,nnscope,(cp,DEST(cp'))) = (replaceVarpsT(index,sub,nnscope,cp),DEST(replaceVarpsT(index,sub,nnscope,cp')))

and replaceVarpsT(index,sub,nnscope,DEFVARP(s)) = (case mapsTo(DEFVARP(s),sub) of NONE => raise Problem("") | SOME(p) => NAME(p))
  | replaceVarpsT(index,sub,nnscope,VAR(s))     = (case mapsTo(VAR(s),sub) of NONE => VAR(s^"''"^index) | SOME(p) => NAME(p))
  | replaceVarpsT(index,sub,nnscope,ENC(ts',key,cpform)) = ENC(replaceVarpsTs(index,sub,nnscope,ts'),replaceVarpsT(index,sub,nnscope,key),replaceVarpsCP(index,sub,nnscope,cpform))
  | replaceVarpsT(index,sub,nnscope,KEY(t))     = KEY(replaceVarpsT(index,sub,nnscope,t))
  | replaceVarpsT(index,sub,nnscope,DEFVAR(s)) = DEFVAR(s^"''"^index)
  | replaceVarpsT(index,sub,nnscope,NAME n) = if Aux.member(n,nnscope) then NAME(n^"''"^index) else NAME n (* index are not added to normal names since they are considered "global"*)

and replaceVarpsTs(index,sub,nnscope,t::ts) = replaceVarpsT(index,sub,nnscope,t)::replaceVarpsTs(index,sub,nnscope,ts)
  | replaceVarpsTs(_,_,_,[]) = []

and replaceVarps(index,sub,nnscope,INP(ts,P)) = INP(replaceVarpsTs(index,sub,nnscope,ts),replaceVarps(index,sub,nnscope,P))
  | replaceVarps(index,sub,nnscope,OUT(ts,P)) = OUT(replaceVarpsTs(index,sub,nnscope,ts),replaceVarps(index,sub,nnscope,P))
  | replaceVarps(index,sub,nnscope,NEW(NAME n,P))  = NEW(NAME (n^"''"^index),replaceVarps(index,sub,n::nnscope,P))
  | replaceVarps(_,_,_,NIL)           = NIL

fun createIndex(p::ps) = p^createIndex(ps) 
  | createIndex([]) = ""
  
(*zip(varps,ps) giver en liste af par der svarer til substitution. denne substitution bliver foretaget i P. Samtidig tilfjes index baseret p den givne substitution til alle new names og variable *)
fun expandP'(varps,ps::pss,P) = let val index = createIndex(ps)
                                in replaceVarps(index,ListPair.zip(varps,ps),[],P) :: expandP'(varps,pss,P) end
  | expandP'(_,[],_)        = []

(*expands the process P to a list of processes where all occurences of principal variables are replaced by names in principals. Furthermore all new (fresh) names and all variables are renamed to avoid clash of names in the new set of processes *)
fun expandP(principals,P) = let val varps = defvarp(P) (*varp list*)
                            in if List.length(varps) = 0 
                               then [P] 
                               else 
                                 let val pss = enumeratePermutations(principals,List.length(varps)) (*list of possible permutations containing n principals, wherre n = length(varps)*)
                                 in expandP'(varps,pss,P) end 
                            end

(*********************************)  

fun convertS'(PAR(Ss),principals) = Lysa.PAR(List.map (fn S => convertS'(S,principals)) Ss)
  | convertS'(LOOP(principal,P),principals) = let val Ps = expandP(principals,P) 
                                              in Lysa.BANG(Lysa.PAR(List.map (fn P => convertP(P)) Ps)) end
  | convertS'(NEWS(n,S),principals) = Lysa.NEW(convertT(n),convertS'(S,principals)) 

fun convertS(lysaN) = (resetVar();convertS'(lysaN))
end
