structure Types =
struct
local
open LysaN
in
type label = string*int

type principal = string

datatype comstep = OUTSTEP of LysaN.Term list
                 | INSTEP  of LysaN.Term list
type transition = principal*comstep

type step = (label * label) * transition (*a transition for one principal from one label to another*)
type flow = step list  (* all possible communication steps for a principal or an entire protocol *)
type state = label list (*combined state of all parallel processes*)

exception Problem of string;

datatype CForm  = CORIG    of CTerm
                | CDEST    of CTerm


and CTerm  = CNAME    of string * int
           | CVAR     of string * int
           | CENC     of CTerm list * CTerm * (CTerm * CForm) * bool (* the bool value is a flag indicating whether we have considered/decrypted the term yet. It only has a purpose when the given term is a part of an IK *)
           | CKEY     of CTerm (*invariant. Term must be a NAME or a principal var*)
(* we do not need defvar and defvarp in the constraints *)

fun formToCForm((cp1,ORIG cp2),termList) = let val ([cp1',cp2'],termList') = termToCTermList([cp1,cp2],termList) in ((cp1',CORIG cp2'),termList') end
  | formToCForm((cp1,DEST cp2),termList) = let val ([cp1',cp2'],termList') = termToCTermList([cp1,cp2],termList) in ((cp1',CDEST cp2'),termList') end

and termToCTerm(NAME n,termList) = let val index = Aux.count(n,termList) in (CNAME(n,index),termList) end
  | termToCTerm(VAR x,termList)  = let val index = Aux.count(x,termList) in (CVAR(x,index),termList) end
  | termToCTerm(DEFVAR(x),termList) =  let val index = Aux.count(x,termList)+1 in (CVAR(x,index),x::termList) end
  | termToCTerm(DEFVARP(x),termList) = let val index = Aux.count(x,termList)+1 in (CVAR(x,index),x::termList) end
  | termToCTerm(ENC(ts,t,cps),termList) = let val (t'::ts',termList') = termToCTermList(t::ts,termList) 
                                              val (cps',termList'') = formToCForm(cps,termList')
                                          in (CENC(ts',t',cps',false),termList'') end
  | termToCTerm(KEY(t),termList) = let val (t',termList') = termToCTerm(t,termList) in (CKEY(t'),termList') end
  | termToCTerm(FRESH(NAME n),termList) = let val index = Aux.count(n,termList)+1 in (CNAME(n,index),n::termList) end


and termToCTermList(t::ts,termList) =  let val (t',termList') = termToCTerm(t,termList) 
                                           val (ts',termList'') = termToCTermList(ts,termList')
                                       in (t'::ts',termList'') end
  | termToCTermList([],termList) = ([],termList)


fun resetCTerm(CENC(ts,t,cps,_)) = CENC(resetCTermList ts,resetCTerm t,cps,false)
  | resetCTerm t = t

and resetCTermList(ts) = List.map resetCTerm ts

(* evaluates to true i names are equal and (index are equal or one of them are zero) *)
fun compareNameAndIndex((n1,i1),(n2,i2)) = case (n1=n2,i1=i2,i1,i2) of
                                            (true,true,_,_) => true
                                           |(true,_,0,_)    => true
                                           |(true,_,_,0)    => true
                                           | _              => false
                                       
infix ==

(* equality function which takes care of index and wildcard and ignores decryption flag*)
fun (CNAME(n1,i1)) == (CNAME(n2,i2)) =   compareNameAndIndex((n1,i1),(n2,i2))
  | (CVAR(x1,i1)) == (CVAR(x2,i2)) =   compareNameAndIndex((x1,i1),(x2,i2))
  | (CENC(ts1,t1,_,_)) == (CENC(ts2,t2,_,_)) = ListPair.all (op ==) (ts1,ts2) andalso t1 == t2
  | (CKEY t1) == (CKEY t2) = t1 == t2
  | _ == _ = false



fun lookUp(t1 as CVAR(_),(t2 as CVAR(_),t)::sub) = if (t1 == t2) then t else lookUp(t1,sub)
  | lookUp(CVAR(x,i),[]) = CVAR(x,i)
  | lookUp _ = raise Problem("lookup")


fun lookUp''(t1 as CVAR(_),(t2 as CVAR(_),t)::sub) = if (t1 <> t2) then lookUp''(t1,sub) else
                                                   (case lookUp''(t1,sub) of
                                                     SOME(CVAR(_)) => SOME t2
                                                    |SOME(t1' as CNAME(_)) => SOME t1'
                                                    |NONE => SOME t2
                                                    )
  | lookUp''(CVAR(x,i),[]) = SOME (CVAR(x,i))
  | lookUp'' _ = raise Problem("lookup")

fun lookUp'(t,sub,lastT) = let val tOpt' = lookUp''(t,sub)
                          in case tOpt' of
                            SOME(CNAME(n,i)) => CNAME(n,i)
                           |SOME(CVAR(n,i)) => 
                              if CVAR(n,i) = lastT then CVAR(n,i) else lookUp'(CVAR(n,i),sub,t)
                          end

(*fun lookUp(t,sub) = lookUp'(t,sub,t)*)

fun applySubstitution sub (t as CVAR(_)) = lookUp(t,sub)
  | applySubstitution sub (t as CNAME(_)) = t
  | applySubstitution sub (CKEY(t)) = CKEY(applySubstitution sub t)
  | applySubstitution sub (CENC(ts,t,(cp,form),b)) = let val ts' = List.map (applySubstitution sub) ts
                                                         val t'  = applySubstitution sub t
                                                         val cp' = applySubstitution sub cp
                                                         val form' = (case form of CORIG(t) => CORIG(applySubstitution sub t) 
                                                                                 | CDEST(t) => CDEST(applySubstitution sub t))
                                                     in CENC(ts',t',(cp',form'),b) end

fun applySubstitutionTs sub ts = List.map (applySubstitution sub) ts


fun intersects(x::xs,ys) = Aux.member(x,ys) orelse intersects(xs,ys)
  | intersects([],ys) = false

fun flatten(CENC(ts,t,(cp,form),b)) = flatten t @ (flatten cp) @ (case form of CORIG(t) => flatten t | CDEST(t) => flatten t) @ Aux.meltList(List.map flatten ts)
 |  flatten(CKEY(t)) = flatten t
 |  flatten(t) = [t]

fun substitutionOk(sub) = let val (dom,range) = ListPair.unzip  sub
                          in not(intersects(dom,Aux.meltList(List.map flatten range))) end

fun fixSubstitution'(sub) = let val sub' = List.map (fn (t1,t2) => (t1,applySubstitution sub t2)) sub
                            in if substitutionOk(sub') then sub' else fixSubstitution'(sub') end

fun fixSubstitution(sub) = let val (s1,s2) = List.partition (fn (t1,t2) => (case t2 of CVAR(_) => true | _ => false)) sub
                           in fixSubstitution'(s1@s2) end


end
end